@@ -14,31 +14,26 @@
 #include "soup-cookie-jar.h"
 #include "soup-date-utils-private.h"
 #include "soup-message-private.h"
+#include "soup-message-headers-private.h"
 #include "soup-misc.h"
 #include "soup.h"
 #include "soup-session-feature-private.h"
 #include "soup-uri-utils-private.h"
 
 /**
- * SECTION:soup-cookie-jar
- * @short_description: Automatic cookie handling for SoupSession
+ * SoupCookieJar:
  *
- * A #SoupCookieJar stores #SoupCookie<!-- -->s and arrange for them
- * to be sent with the appropriate #SoupMessage<!-- -->s.
- * #SoupCookieJar implements #SoupSessionFeature, so you can add a
- * cookie jar to a session with soup_session_add_feature() or
- * soup_session_add_feature_by_type().
+ * Automatic cookie handling for SoupSession.
+ *
+ * A #SoupCookieJar stores [struct@Cookie]s and arrange for them to be sent with
+ * the appropriate [class@Message]s. #SoupCookieJar implements
+ * [iface@SessionFeature], so you can add a cookie jar to a session with
+ * [method@Session.add_feature] or [method@Session.add_feature_by_type].
  *
  * Note that the base #SoupCookieJar class does not support any form
  * of long-term cookie persistence.
  **/
 
-/**
- * SoupCookieJar:
- *
- * Class that stores cookies in memory.
- */
-
 enum {
 	CHANGED,
 	LAST_SIGNAL
@@ -58,6 +53,7 @@ enum {
 static GParamSpec *properties[LAST_PROPERTY] = { NULL, };
 
 typedef struct {
+        GMutex mutex;
 	gboolean constructed, read_only;
 	GHashTable *domains, *serials;
 	guint serial;
@@ -81,6 +77,7 @@ soup_cookie_jar_init (SoupCookieJar *jar)
 					       g_free, NULL);
 	priv->serials = g_hash_table_new (NULL, NULL);
 	priv->accept_policy = SOUP_COOKIE_JAR_ACCEPT_ALWAYS;
+        g_mutex_init (&priv->mutex);
 }
 
 static void
@@ -105,6 +102,7 @@ soup_cookie_jar_finalize (GObject *object)
 		soup_cookies_free (value);
 	g_hash_table_destroy (priv->domains);
 	g_hash_table_destroy (priv->serials);
+        g_mutex_clear (&priv->mutex);
 
 	G_OBJECT_CLASS (soup_cookie_jar_parent_class)->finalize (object);
 }
@@ -173,7 +171,9 @@ soup_cookie_jar_class_init (SoupCookieJarClass *jar_class)
 	 * @old_cookie: the old #SoupCookie value
 	 * @new_cookie: the new #SoupCookie value
 	 *
-	 * Emitted when @jar changes. If a cookie has been added,
+	 * Emitted when @jar changes.
+	 *
+	 * If a cookie has been added,
 	 * @new_cookie will contain the newly-added cookie and
 	 * @old_cookie will be %NULL. If a cookie has been deleted,
 	 * @old_cookie will contain the to-be-deleted cookie and
@@ -192,6 +192,11 @@ soup_cookie_jar_class_init (SoupCookieJarClass *jar_class)
 			      SOUP_TYPE_COOKIE | G_SIGNAL_TYPE_STATIC_SCOPE,
 			      SOUP_TYPE_COOKIE | G_SIGNAL_TYPE_STATIC_SCOPE);
 
+	/**
+	 * SoupCookieJar:read-only:
+	 *
+	 * Whether or not the cookie jar is read-only.
+	 */
         properties[PROP_READ_ONLY] =
 		g_param_spec_boolean ("read-only",
 				      "Read-only",
@@ -201,10 +206,9 @@ soup_cookie_jar_class_init (SoupCookieJarClass *jar_class)
 				      G_PARAM_STATIC_STRINGS);
 
 	/**
-	 * SoupCookieJar:accept-policy:
-	 *
-	 * The policy the jar should follow to accept or reject cookies
+	 * SoupCookieJar:accept-policy: (attributes org.gtk.Property.get=soup_cookie_jar_get_accept_policy org.gtk.Property.set=soup_cookie_jar_set_accept_policy)
 	 *
+	 * The policy the jar should follow to accept or reject cookies.
 	 */
         properties[PROP_ACCEPT_POLICY] =
 		g_param_spec_enum ("accept-policy",
@@ -221,11 +225,12 @@ soup_cookie_jar_class_init (SoupCookieJarClass *jar_class)
 /**
  * soup_cookie_jar_new:
  *
- * Creates a new #SoupCookieJar. The base #SoupCookieJar class does
- * not support persistent storage of cookies; use a subclass for that.
+ * Creates a new #SoupCookieJar.
+ *
+ * The base #SoupCookieJar class does not support persistent storage of cookies;
+ * use a subclass for that.
  *
  * Returns: a new #SoupCookieJar
- *
  **/
 SoupCookieJar *
 soup_cookie_jar_new (void) 
@@ -340,6 +345,8 @@ get_cookies (SoupCookieJar *jar,
                 next_domain = NULL;
         }
 
+        g_mutex_lock (&priv->mutex);
+
 	do {
 		new_head = domain_cookies = g_hash_table_lookup (priv->domains, cur);
 		while (domain_cookies) {
@@ -376,6 +383,8 @@ get_cookies (SoupCookieJar *jar,
 	}
 	g_slist_free (cookies_to_remove);
 
+        g_mutex_unlock (&priv->mutex);
+
 	return g_slist_sort_with_data (cookies, compare_cookies, jar);
 }
 
@@ -384,7 +393,7 @@ get_cookies (SoupCookieJar *jar,
  * @jar: a #SoupCookieJar
  * @uri: a #GUri
  * @for_http: whether or not the return value is being passed directly
- * to an HTTP operation
+ *   to an HTTP operation
  *
  * Retrieves (in Cookie-header form) the list of cookies that would
  * be sent with a request to @uri.
@@ -398,8 +407,7 @@ get_cookies (SoupCookieJar *jar,
  * this.
  *
  * Returns: (nullable): the cookies, in string form, or %NULL if
- * there are no cookies for @uri.
- *
+ *   there are no cookies for @uri.
  **/
 char *
 soup_cookie_jar_get_cookies (SoupCookieJar *jar, GUri *uri,
@@ -430,10 +438,10 @@ soup_cookie_jar_get_cookies (SoupCookieJar *jar, GUri *uri,
  * @jar: a #SoupCookieJar
  * @uri: a #GUri
  * @for_http: whether or not the return value is being passed directly
- * to an HTTP operation
+ *   to an HTTP operation
  *
  * Retrieves the list of cookies that would be sent with a request to @uri
- * as a #GSList of #SoupCookie objects.
+ * as a [struct@GLib.List] of #SoupCookie objects.
  *
  * If @for_http is %TRUE, the return value will include cookies marked
  * "HttpOnly" (that is, cookies that the server wishes to keep hidden
@@ -444,8 +452,7 @@ soup_cookie_jar_get_cookies (SoupCookieJar *jar, GUri *uri,
  * this.
  *
  * Returns: (transfer full) (element-type Soup.Cookie): a #GSList
- * with the cookies in the @jar that would be sent with a request to @uri.
- *
+ *   with the cookies in the @jar that would be sent with a request to @uri.
  **/
 GSList *
 soup_cookie_jar_get_cookie_list (SoupCookieJar *jar, GUri *uri, gboolean for_http)
@@ -463,19 +470,20 @@ soup_cookie_jar_get_cookie_list (SoupCookieJar *jar, GUri *uri, gboolean for_htt
  * @top_level: (nullable): a #GUri for the top level document
  * @site_for_cookies: (nullable): a #GUri indicating the origin to get cookies for
  * @for_http: whether or not the return value is being passed directly
- * to an HTTP operation
+ *   to an HTTP operation
  * @is_safe_method: if the HTTP method is safe, as defined by RFC 7231, ignored when @for_http is %FALSE
  * @is_top_level_navigation: whether or not the HTTP request is part of
- * top level navigation
+ *   top level navigation
  *
- * This is an extended version of soup_cookie_jar_get_cookie_list() that
- * provides more information required to use SameSite cookies. See the
- * [SameSite cookies spec](https://tools.ietf.org/html/draft-ietf-httpbis-cookie-same-site-00)
- * for more detailed information.
+ * This is an extended version of [method@CookieJar.get_cookie_list] that
+ * provides more information required to use SameSite cookies.
+ *
+ * See the [SameSite cookies
+ * spec](https://tools.ietf.org/html/draft-ietf-httpbis-cookie-same-site-00) for
+ * more detailed information.
  *
  * Returns: (transfer full) (element-type Soup.Cookie): a #GSList
- * with the cookies in the @jar that would be sent with a request to @uri.
- *
+ *   with the cookies in the @jar that would be sent with a request to @uri.
  */
 GSList *
 soup_cookie_jar_get_cookie_list_with_same_site_info (SoupCookieJar *jar,
@@ -514,6 +522,7 @@ incoming_cookie_is_third_party (SoupCookieJar            *jar,
 	const char *cookie_base_domain;
 	const char *first_party_base_domain;
         const char *first_party_host;
+        gboolean retval;
 
 	if (policy != SOUP_COOKIE_JAR_ACCEPT_NO_THIRD_PARTY &&
 	    policy != SOUP_COOKIE_JAR_ACCEPT_GRANDFATHERED_THIRD_PARTY)
@@ -548,7 +557,26 @@ incoming_cookie_is_third_party (SoupCookieJar            *jar,
 	 * previously visited directly.
 	 */
 	priv = soup_cookie_jar_get_instance_private (jar);
-	return !g_hash_table_lookup (priv->domains, soup_cookie_get_domain (cookie));
+        g_mutex_lock (&priv->mutex);
+	retval = !g_hash_table_lookup (priv->domains, soup_cookie_get_domain (cookie));
+        g_mutex_unlock (&priv->mutex);
+
+        return retval;
+}
+
+static gboolean
+string_contains_ctrlcode (const char *s)
+{
+	const char *p;
+
+	p = s;
+	while (*p != '\0') {
+		if (g_ascii_iscntrl (*p) && *p != 0x09)
+			return TRUE;
+		
+		p++;
+	}
+	return FALSE;
 }
 
 /**
@@ -558,9 +586,11 @@ incoming_cookie_is_third_party (SoupCookieJar            *jar,
  * @uri: (nullable): the URI setting the cookie
  * @first_party: (nullable): the URI for the main document
  *
- * Adds @cookie to @jar, emitting the 'changed' signal if we are modifying
- * an existing cookie or adding a valid new cookie ('valid' means
- * that the cookie's expire date is not in the past).
+ * Adds @cookie to @jar.
+ *
+ * Emits the [signal@CookieJar::changed] signal if we are modifying an existing
+ * cookie or adding a valid new cookie ('valid' means that the cookie's expire
+ * date is not in the past).
  *
  * @first_party will be used to reject cookies coming from third party
  * resources in case such a security policy is set in the @jar.
@@ -569,7 +599,6 @@ incoming_cookie_is_third_party (SoupCookieJar            *jar,
  * from insecure origins. %NULL is treated as secure.
  * 
  * @cookie will be 'stolen' by the jar, so don't free it afterwards.
- *
  **/
 void
 soup_cookie_jar_add_cookie_full (SoupCookieJar *jar, SoupCookie *cookie, GUri *uri, GUri *first_party)
@@ -604,6 +633,50 @@ soup_cookie_jar_add_cookie_full (SoupCookieJar *jar, SoupCookie *cookie, GUri *u
 		return;
 	}
 
+	/* SameSite=None cookies are rejected unless the Secure attribute is set. */
+	if (soup_cookie_get_same_site_policy (cookie) == SOUP_SAME_SITE_POLICY_NONE && !soup_cookie_get_secure (cookie)) {
+		soup_cookie_free (cookie);
+		return;
+	}
+
+        /* See https://datatracker.ietf.org/doc/html/draft-ietf-httpbis-cookie-prefixes-00 for handling the prefixes,
+         * which has been implemented by Firefox and Chrome. */
+#define MATCH_PREFIX(name, prefix) (!g_ascii_strncasecmp (name, prefix, strlen(prefix)))
+
+	/* Cookies with a "__Secure-" prefix should have Secure attribute set and it must be for a secure host. */
+	if (MATCH_PREFIX (soup_cookie_get_name (cookie), "__Secure-") && (!soup_cookie_get_secure (cookie) || !uri)) {
+		soup_cookie_free (cookie);
+		return;
+	}
+        /* Path=/ and Secure attributes are required; Domain attribute must not be present.
+         Note that SoupCookie always sets the domain so we do exact host matches instead of subdomain matches. */
+	if (MATCH_PREFIX (soup_cookie_get_name (cookie), "__Host-")) {
+		if ((!soup_cookie_get_secure (cookie) || !uri) ||
+		    strcmp (soup_cookie_get_path (cookie), "/") != 0 ||
+		    g_ascii_strcasecmp (soup_cookie_get_domain (cookie), g_uri_get_host (uri)) != 0) {
+			soup_cookie_free (cookie);
+			return;
+		}
+	}
+
+	/* Cookies should not take control characters %x00-1F / %x7F (defined by RFC 5234) in names or values,
+	 * with the exception of %x09 (the tab character).
+	 */    
+	const char *name, *value;
+	name = soup_cookie_get_name (cookie);
+	value = soup_cookie_get_value (cookie);
+	if (string_contains_ctrlcode (name) || string_contains_ctrlcode (value)) {
+		soup_cookie_free (cookie);
+		return;
+	}
+	
+	if (strlen(name) > 4096 || strlen(value) > 4096) {
+		soup_cookie_free (cookie);
+		return;
+	}
+	
+        g_mutex_lock (&priv->mutex);
+
 	old_cookies = g_hash_table_lookup (priv->domains, soup_cookie_get_domain (cookie));
 	for (oc = old_cookies; oc; oc = oc->next) {
 		old_cookie = oc->data;
@@ -633,6 +706,8 @@ soup_cookie_jar_add_cookie_full (SoupCookieJar *jar, SoupCookie *cookie, GUri *u
 				soup_cookie_free (old_cookie);
 			}
 
+                        g_mutex_unlock (&priv->mutex);
+
 			return;
 		}
 		last = oc;
@@ -641,6 +716,7 @@ soup_cookie_jar_add_cookie_full (SoupCookieJar *jar, SoupCookie *cookie, GUri *u
 	/* The new cookie is... a new cookie */
 	if (soup_cookie_get_expires (cookie) && soup_date_time_is_past (soup_cookie_get_expires (cookie))) {
 		soup_cookie_free (cookie);
+                g_mutex_unlock (&priv->mutex);
 		return;
 	}
 
@@ -653,6 +729,8 @@ soup_cookie_jar_add_cookie_full (SoupCookieJar *jar, SoupCookie *cookie, GUri *u
 	}
 
 	soup_cookie_jar_changed (jar, NULL, cookie);
+
+        g_mutex_unlock (&priv->mutex);
 }
 
 /**
@@ -660,12 +738,13 @@ soup_cookie_jar_add_cookie_full (SoupCookieJar *jar, SoupCookie *cookie, GUri *u
  * @jar: a #SoupCookieJar
  * @cookie: (transfer full): a #SoupCookie
  *
- * Adds @cookie to @jar, emitting the 'changed' signal if we are modifying
+ * Adds @cookie to @jar.
+ *
+ * Emits the [signal@CookieJar::changed] signal if we are modifying
  * an existing cookie or adding a valid new cookie ('valid' means
  * that the cookie's expire date is not in the past).
  *
  * @cookie will be 'stolen' by the jar, so don't free it afterwards.
- *
  **/
 void
 soup_cookie_jar_add_cookie (SoupCookieJar *jar, SoupCookie *cookie)
@@ -679,7 +758,9 @@ soup_cookie_jar_add_cookie (SoupCookieJar *jar, SoupCookie *cookie)
  * @first_party: the URI for the main document
  * @cookie: (transfer full): a #SoupCookie
  *
- * Adds @cookie to @jar, emitting the 'changed' signal if we are modifying
+ * Adds @cookie to @jar.
+ *
+ * Emits the [signal@CookieJar::changed] signal if we are modifying
  * an existing cookie or adding a valid new cookie ('valid' means
  * that the cookie's expire date is not in the past).
  *
@@ -689,8 +770,7 @@ soup_cookie_jar_add_cookie (SoupCookieJar *jar, SoupCookie *cookie)
  * @cookie will be 'stolen' by the jar, so don't free it afterwards.
  *
  * For secure cookies to work properly you may want to use
- * soup_cookie_jar_add_cookie_full().
- *
+ * [method@CookieJar.add_cookie_full].
  **/
 void
 soup_cookie_jar_add_cookie_with_first_party (SoupCookieJar *jar, GUri *first_party, SoupCookie *cookie)
@@ -709,13 +789,12 @@ soup_cookie_jar_add_cookie_with_first_party (SoupCookieJar *jar, GUri *first_par
  * Adds @cookie to @jar, exactly as though it had appeared in a
  * Set-Cookie header returned from a request to @uri.
  *
- * Keep in mind that if the #SoupCookieJarAcceptPolicy set is either
+ * Keep in mind that if the [enum@CookieJarAcceptPolicy] set is either
  * %SOUP_COOKIE_JAR_ACCEPT_NO_THIRD_PARTY or
  * %SOUP_COOKIE_JAR_ACCEPT_GRANDFATHERED_THIRD_PARTY you'll need to use
- * soup_cookie_jar_set_cookie_with_first_party(), otherwise the jar
+ * [method@CookieJar.set_cookie_with_first_party], otherwise the jar
  * will have no way of knowing if the cookie is being set by a third
  * party or not.
- *
  **/
 void
 soup_cookie_jar_set_cookie (SoupCookieJar *jar, GUri *uri,
@@ -753,10 +832,10 @@ soup_cookie_jar_set_cookie (SoupCookieJar *jar, GUri *uri,
  * @cookie: the stringified cookie to set
  *
  * Adds @cookie to @jar, exactly as though it had appeared in a
- * Set-Cookie header returned from a request to @uri. @first_party
- * will be used to reject cookies coming from third party resources in
- * case such a security policy is set in the @jar.
+ * Set-Cookie header returned from a request to @uri.
  *
+ * @first_party will be used to reject cookies coming from third party resources
+ * in case such a security policy is set in the @jar.
  **/
 void
 soup_cookie_jar_set_cookie_with_first_party (SoupCookieJar *jar,
@@ -814,11 +893,11 @@ msg_starting_cb (SoupMessage *msg, gpointer feature)
 							               soup_message_get_is_top_level_navigation (msg));
 	if (cookies != NULL) {
 		char *cookie_header = soup_cookies_to_cookie_header (cookies);
-		soup_message_headers_replace (soup_message_get_request_headers (msg), "Cookie", cookie_header);
+		soup_message_headers_replace_common (soup_message_get_request_headers (msg), SOUP_HEADER_COOKIE, cookie_header);
 		g_free (cookie_header);
 		g_slist_free_full (cookies, (GDestroyNotify)soup_cookie_free);
 	} else {
-		soup_message_headers_remove (soup_message_get_request_headers (msg), "Cookie");
+		soup_message_headers_remove_common (soup_message_get_request_headers (msg), SOUP_HEADER_COOKIE);
 	}
 }
 
@@ -859,13 +938,13 @@ soup_cookie_jar_session_feature_init (SoupSessionFeatureInterface *feature_inter
  * soup_cookie_jar_all_cookies:
  * @jar: a #SoupCookieJar
  *
- * Constructs a #GSList with every cookie inside the @jar.
+ * Constructs a [struct@GLib.List] with every cookie inside the @jar.
+ *
  * The cookies in the list are a copy of the original, so
  * you have to free them when you are done with them.
  *
  * Returns: (transfer full) (element-type Soup.Cookie): a #GSList
- * with all the cookies in the @jar.
- *
+ *   with all the cookies in the @jar.
  **/
 GSList *
 soup_cookie_jar_all_cookies (SoupCookieJar *jar)
@@ -879,6 +958,8 @@ soup_cookie_jar_all_cookies (SoupCookieJar *jar)
 
 	priv = soup_cookie_jar_get_instance_private (jar);
 
+        g_mutex_lock (&priv->mutex);
+
 	g_hash_table_iter_init (&iter, priv->domains);
 
 	while (g_hash_table_iter_next (&iter, &key, &value)) {
@@ -887,6 +968,8 @@ soup_cookie_jar_all_cookies (SoupCookieJar *jar)
 			l = g_slist_prepend (l, soup_cookie_copy (p->data));
 	}
 
+        g_mutex_unlock (&priv->mutex);
+
 	return l;
 }
 
@@ -895,8 +978,9 @@ soup_cookie_jar_all_cookies (SoupCookieJar *jar)
  * @jar: a #SoupCookieJar
  * @cookie: a #SoupCookie
  *
- * Deletes @cookie from @jar, emitting the 'changed' signal.
+ * Deletes @cookie from @jar.
  *
+ * Emits the [signal@CookieJar::changed] signal.
  **/
 void
 soup_cookie_jar_delete_cookie (SoupCookieJar *jar,
@@ -910,9 +994,13 @@ soup_cookie_jar_delete_cookie (SoupCookieJar *jar,
 
 	priv = soup_cookie_jar_get_instance_private (jar);
 
+        g_mutex_lock (&priv->mutex);
+
 	cookies = g_hash_table_lookup (priv->domains, soup_cookie_get_domain (cookie));
-	if (cookies == NULL)
+	if (cookies == NULL) {
+                g_mutex_unlock (&priv->mutex);
 		return;
+        }
 
 	for (p = cookies; p; p = p->next ) {
 		SoupCookie *c = (SoupCookie*)p->data;
@@ -923,53 +1011,54 @@ soup_cookie_jar_delete_cookie (SoupCookieJar *jar,
 					     cookies);
 			soup_cookie_jar_changed (jar, c, NULL);
 			soup_cookie_free (c);
+                        g_mutex_unlock (&priv->mutex);
 			return;
 		}
 	}
+
+        g_mutex_unlock (&priv->mutex);
 }
 
 /**
  * SoupCookieJarAcceptPolicy:
  * @SOUP_COOKIE_JAR_ACCEPT_ALWAYS: accept all cookies unconditionally.
  * @SOUP_COOKIE_JAR_ACCEPT_NEVER: reject all cookies unconditionally.
- * @SOUP_COOKIE_JAR_ACCEPT_NO_THIRD_PARTY: accept all cookies set by
- * the main document loaded in the application using libsoup. An
- * example of the most common case, web browsers, would be: If
- * http://www.example.com is the page loaded, accept all cookies set
- * by example.com, but if a resource from http://www.third-party.com
- * is loaded from that page reject any cookie that it could try to
- * set. For libsoup to be able to tell apart first party cookies from
- * the rest, the application must call soup_message_set_first_party()
- * on each outgoing #SoupMessage, setting the #GUri of the main
- * document. If no first party is set in a message when this policy is
- * in effect, cookies will be assumed to be third party by default.
- * @SOUP_COOKIE_JAR_ACCEPT_GRANDFATHERED_THIRD_PARTY: accept all cookies
- * set by the main document loaded in the application using libsoup, and
- * from domains that have previously set at least one cookie when loaded
- * as the main document. An example of the most common case, web browsers,
- * would be: if http://www.example.com is the page loaded, accept all
- * cookies set by example.com, but if a resource from http://www.third-party.com
- * is loaded from that page, reject any cookie that it could try to
- * set unless it already has a cookie in the cookie jar. For libsoup to
- * be able to tell apart first party cookies from the rest, the
- * application must call soup_message_set_first_party() on each outgoing
- * #SoupMessage, setting the #GUri of the main document. If no first
- * party is set in a message when this policy is in effect, cookies will
- * be assumed to be third party by default.
+ * @SOUP_COOKIE_JAR_ACCEPT_NO_THIRD_PARTY: accept all cookies set by the main
+ *   document loaded in the application using libsoup. An example of the most
+ *   common case, web browsers, would be: If http://www.example.com is the page
+ *   loaded, accept all cookies set by example.com, but if a resource from
+ *   http://www.third-party.com is loaded from that page reject any cookie that
+ *   it could try to set. For libsoup to be able to tell apart first party
+ *   cookies from the rest, the application must call
+ *   [method@Message.set_first_party] on each outgoing [class@Message], setting
+ *   the [struct@GLib.Uri] of the main document. If no first party is set in a
+ *   message when this policy is in effect, cookies will be assumed to be third
+ *   party by default.
+ * @SOUP_COOKIE_JAR_ACCEPT_GRANDFATHERED_THIRD_PARTY: accept all cookies set by
+ *   the main document loaded in the application using libsoup, and from domains
+ *   that have previously set at least one cookie when loaded as the main
+ *   document. An example of the most common case, web browsers, would be: if
+ *   http://www.example.com is the page loaded, accept all cookies set by
+ *   example.com, but if a resource from http://www.third-party.com is loaded
+ *   from that page, reject any cookie that it could try to set unless it
+ *   already has a cookie in the cookie jar. For libsoup to be able to tell
+ *   apart first party cookies from the rest, the application must call
+ *   [method@Message.set_first_party] on each outgoing #SoupMessage, setting the
+ *   [struct@GLib.Uri] of the main document. If no first party is set in a
+ *   message when this policy is in effect, cookies will be assumed to be third
+ *   party by default.
  *
  * The policy for accepting or rejecting cookies returned in
  * responses.
- *
  */
 
 /**
- * soup_cookie_jar_get_accept_policy:
+ * soup_cookie_jar_get_accept_policy: (attributes org.gtk.Method.get_property=accept-policy)
  * @jar: a #SoupCookieJar
  *
- * Gets @jar's #SoupCookieJarAcceptPolicy
+ * Gets @jar's [enum@CookieJarAcceptPolicy].
  *
  * Returns: the #SoupCookieJarAcceptPolicy set in the @jar
- *
  **/
 SoupCookieJarAcceptPolicy
 soup_cookie_jar_get_accept_policy (SoupCookieJar *jar)
@@ -983,12 +1072,11 @@ soup_cookie_jar_get_accept_policy (SoupCookieJar *jar)
 }
 
 /**
- * soup_cookie_jar_set_accept_policy:
+ * soup_cookie_jar_set_accept_policy: (attributes org.gtk.Method.set_property=accept-policy)
  * @jar: a #SoupCookieJar
  * @policy: a #SoupCookieJarAcceptPolicy
  * 
  * Sets @policy as the cookie acceptance policy for @jar.
- *
  **/
 void
 soup_cookie_jar_set_accept_policy (SoupCookieJar *jar,
@@ -1013,7 +1101,6 @@ soup_cookie_jar_set_accept_policy (SoupCookieJar *jar,
  * Gets whether @jar stores cookies persistenly.
  *
  * Returns: %TRUE if @jar storage is persistent or %FALSE otherwise.
- *
  **/
 gboolean
 soup_cookie_jar_is_persistent (SoupCookieJar *jar)
