@@ -66,6 +66,8 @@ typedef struct {
         gpointer started_user_data;
 
         GHashTable *messages;
+
+        guint in_callback;
 } SoupServerMessageIOHTTP2;
 
 static void soup_server_message_io_http2_send_response (SoupServerMessageIOHTTP2 *io,
@@ -324,6 +326,7 @@ io_write (SoupServerMessageIOHTTP2 *io,
 
         if (io->write_buffer == NULL) {
                 io->written_bytes = 0;
+                g_assert (io->in_callback == 0);
                 io->write_buffer_size = nghttp2_session_mem_send (io->session, (const guint8**)&io->write_buffer);
                 if (io->write_buffer_size == 0) {
                         /* Done */
@@ -347,18 +350,32 @@ static gboolean
 io_write_ready (GObject                  *stream,
                 SoupServerMessageIOHTTP2 *io)
 {
+        SoupServerConnection *conn = io->conn;
         GError *error = NULL;
 
-        while (nghttp2_session_want_write (io->session) && !error)
+        g_object_ref (conn);
+
+        while (!error && soup_server_connection_get_io_data (conn) == (SoupServerMessageIO *)io && nghttp2_session_want_write (io->session))
                 io_write (io, &error);
 
         if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
                 g_error_free (error);
+                g_object_unref (conn);
                 return G_SOURCE_CONTINUE;
         }
 
+        if (soup_server_connection_get_io_data (conn) == (SoupServerMessageIO *)io) {
+                if (error)
+                        h2_debug (io, NULL, "[SESSION] IO error: %s", error->message);
+
+                g_clear_pointer (&io->write_source, g_source_unref);
+
+                if (error || (!nghttp2_session_want_read (io->session) && !nghttp2_session_want_write (io->session)))
+                        soup_server_connection_disconnect (conn);
+        }
+
         g_clear_error (&error);
-        g_clear_pointer (&io->write_source, g_source_unref);
+        g_object_unref (conn);
 
         return G_SOURCE_REMOVE;
 }
@@ -366,23 +383,35 @@ io_write_ready (GObject                  *stream,
 static void
 io_try_write (SoupServerMessageIOHTTP2 *io)
 {
+        SoupServerConnection *conn = io->conn;
         GError *error = NULL;
 
         if (io->write_source)
                 return;
 
-        while (nghttp2_session_want_write (io->session) && !error)
+        g_object_ref (conn);
+
+        while (!error && soup_server_connection_get_io_data (conn) == (SoupServerMessageIO *)io && !io->in_callback && nghttp2_session_want_write (io->session))
                 io_write (io, &error);
 
-        if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
-                g_clear_error (&error);
-                io->write_source = g_pollable_output_stream_create_source (G_POLLABLE_OUTPUT_STREAM (io->ostream), NULL);
-                g_source_set_name (io->write_source, "Soup server HTTP/2 write source");
-                g_source_set_callback (io->write_source, (GSourceFunc)io_write_ready, io, NULL);
-                g_source_attach (io->write_source, g_main_context_get_thread_default ());
+        if (soup_server_connection_get_io_data (conn) == (SoupServerMessageIO *)io) {
+                if (io->in_callback || g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
+                        g_clear_error (&error);
+                        io->write_source = g_pollable_output_stream_create_source (G_POLLABLE_OUTPUT_STREAM (io->ostream), NULL);
+                        g_source_set_name (io->write_source, "S oup server HTTP/2 write source");
+                        g_source_set_callback (io->write_source, (GSourceFunc)io_write_ready, io, NULL);
+                        g_source_attach (io->write_source, g_main_context_get_thread_default ());
+                }
+
+                if (error)
+                        h2_debug (io, NULL, "[SESSION] IO error: %s", error->message);
+
+                if (error || (!nghttp2_session_want_read (io->session) && !nghttp2_session_want_write (io->session)))
+                        soup_server_connection_disconnect (conn);
         }
 
         g_clear_error (&error);
+        g_object_unref (conn);
 }
 
 static gboolean
@@ -391,29 +420,57 @@ io_read (SoupServerMessageIOHTTP2 *io,
 {
         guint8 buffer[8192];
         gssize read;
+        int ret;
 
         if ((read = g_pollable_stream_read (io->istream, buffer, sizeof (buffer), FALSE, NULL, error)) < 0)
                 return FALSE;
 
-        return nghttp2_session_mem_recv (io->session, buffer, read) != 0;
+        if (read == 0) {
+                g_set_error_literal (error, G_IO_ERROR,
+                                     G_IO_ERROR_PARTIAL_INPUT,
+                                     _("Connection terminated unexpectedly"));
+                return FALSE;
+        }
+
+        g_assert (io->in_callback == 0);
+        ret = nghttp2_session_mem_recv (io->session, buffer, read);
+        if (ret < 0) {
+                g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "HTTP/2 IO error: %s", nghttp2_strerror (ret));
+                return FALSE;
+        }
+
+        return TRUE;
 }
 
 static gboolean
 io_read_ready (GObject                  *stream,
                SoupServerMessageIOHTTP2 *io)
 {
+        SoupServerConnection *conn = io->conn;
         gboolean progress = TRUE;
         GError *error = NULL;
 
-        while (nghttp2_session_want_read (io->session) && progress)
+        g_object_ref (conn);
+
+        while (progress && soup_server_connection_get_io_data (conn) == (SoupServerMessageIO *)io && nghttp2_session_want_read (io->session))
                 progress = io_read (io, &error);
 
         if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
                 g_error_free (error);
+                g_object_unref (conn);
                 return G_SOURCE_CONTINUE;
         }
 
+        if (soup_server_connection_get_io_data (conn) == (SoupServerMessageIO *)io) {
+                if (error)
+                        h2_debug (io, NULL, "[SESSION] IO error: %s", error->message);
+
+                if (error || (!nghttp2_session_want_read (io->session) && !nghttp2_session_want_write (io->session)))
+                        soup_server_connection_disconnect (conn);
+        }
+
         g_clear_error (&error);
+        g_object_unref (conn);
 
         return G_SOURCE_REMOVE;
 }
@@ -459,6 +516,8 @@ on_begin_headers_callback (nghttp2_session     *session,
         if (frame->hd.type != NGHTTP2_HEADERS || frame->headers.cat != NGHTTP2_HCAT_REQUEST)
                 return 0;
 
+        io->in_callback++;
+
         msg_io = soup_server_message_io_http2_get_or_create_msg_io (io, frame->hd.stream_id);
         h2_debug (io, msg_io, "[SESSION] Message IO created");
 
@@ -469,6 +528,8 @@ on_begin_headers_callback (nghttp2_session     *session,
 
         advance_state_from (msg_io, STATE_NONE, STATE_READ_HEADERS);
 
+        io->in_callback--;
+
         return 0;
 }
 
@@ -482,6 +543,7 @@ on_header_callback (nghttp2_session     *session,
                     uint8_t              flags,
                     void                *user_data)
 {
+        SoupServerMessageIOHTTP2 *io = (SoupServerMessageIOHTTP2 *)user_data;
         SoupMessageIOHTTP2 *msg_io;
         SoupServerMessage *msg;
 
@@ -495,6 +557,8 @@ on_header_callback (nghttp2_session     *session,
         if (!msg_io)
                 return 0;
 
+        io->in_callback++;
+
         msg = msg_io->msg;
         if (name[0] == ':') {
                 if (strcmp ((char *)name, ":method") == 0)
@@ -507,11 +571,14 @@ on_header_callback (nghttp2_session     *session,
                         msg_io->path = g_strndup ((char *)value, valuelen);
                 else
                         g_debug ("Unknown header: %s = %s", name, value);
+                io->in_callback--;
                 return 0;
         }
 
         soup_message_headers_append_untrusted_data (soup_server_message_get_request_headers (msg),
                                                     (const char*)name, (const char*)value);
+
+        io->in_callback--;
         return 0;
 }
 
@@ -523,6 +590,7 @@ on_data_chunk_recv_callback (nghttp2_session *session,
                              size_t           len,
                              void            *user_data)
 {
+        SoupServerMessageIOHTTP2 *io = (SoupServerMessageIOHTTP2 *)user_data;
         SoupMessageIOHTTP2 *msg_io;
         GBytes *bytes;
 
@@ -532,11 +600,15 @@ on_data_chunk_recv_callback (nghttp2_session *session,
 
         h2_debug (user_data, msg_io, "[DATA] Received chunk, len=%zu, flags=%u, paused=%d", len, flags, msg_io->paused);
 
+        io->in_callback++;
+
         bytes = g_bytes_new (data, len);
         soup_message_body_got_chunk (soup_server_message_get_request_body (msg_io->msg), bytes);
         soup_server_message_got_chunk (msg_io->msg, bytes);
         g_bytes_unref (bytes);
 
+        io->in_callback--;
+
         return 0;
 }
 
@@ -549,10 +621,13 @@ on_data_source_read_callback (nghttp2_session     *session,
                               nghttp2_data_source *source,
                               void                *user_data)
 {
+        SoupServerMessageIOHTTP2 *io = (SoupServerMessageIOHTTP2 *)user_data;
         SoupMessageIOHTTP2 *msg_io;
         gsize bytes_written = 0;
         SoupMessageBody *response_body = (SoupMessageBody *)source->ptr;
 
+        io->in_callback++;
+
         msg_io = nghttp2_session_get_stream_user_data (session, stream_id);
 
         h2_debug (user_data, msg_io, "[SEND_BODY] paused=%d", msg_io->paused);
@@ -588,6 +663,8 @@ on_data_source_read_callback (nghttp2_session     *session,
                 *data_flags |= NGHTTP2_DATA_FLAG_EOF;
         }
 
+        io->in_callback--;
+
         return bytes_written;
 }
 
@@ -652,6 +729,8 @@ on_frame_recv_callback (nghttp2_session     *session,
         if (!msg_io)
                 return 0;
 
+        io->in_callback++;
+
         switch (frame->hd.type) {
         case NGHTTP2_HEADERS: {
                 char *uri_string;
@@ -670,6 +749,7 @@ on_frame_recv_callback (nghttp2_session     *session,
         case NGHTTP2_DATA:
                 break;
         default:
+                io->in_callback--;
                 return 0;
         }
 
@@ -679,6 +759,7 @@ on_frame_recv_callback (nghttp2_session     *session,
                 soup_server_message_io_http2_send_response (io, msg_io);
         }
 
+        io->in_callback--;
         return 0;
 }
 
@@ -687,8 +768,11 @@ on_frame_send_callback (nghttp2_session     *session,
                         const nghttp2_frame *frame,
                         void                *user_data)
 {
+        SoupServerMessageIOHTTP2 *io = (SoupServerMessageIOHTTP2 *)user_data;
         SoupMessageIOHTTP2 *msg_io;
 
+        io->in_callback++;
+
         msg_io = nghttp2_session_get_stream_user_data (session, frame->hd.stream_id);
 
         h2_debug (user_data, msg_io, "[SEND] [%s]", soup_http2_frame_type_to_string (frame->hd.type));
@@ -710,6 +794,7 @@ on_frame_send_callback (nghttp2_session     *session,
                 break;
         }
 
+        io->in_callback--;
         return 0;
 }
 
@@ -719,6 +804,7 @@ on_stream_close_callback (nghttp2_session *session,
                           uint32_t         error_code,
                           void            *user_data)
 {
+        SoupServerMessageIOHTTP2 *io = (SoupServerMessageIOHTTP2 *)user_data;
         SoupMessageIOHTTP2 *msg_io;
 
         msg_io = nghttp2_session_get_stream_user_data (session, stream_id);
@@ -726,9 +812,12 @@ on_stream_close_callback (nghttp2_session *session,
         if (!msg_io)
                 return 0;
 
+        io->in_callback++;
+
         if (!msg_io->paused)
                 soup_server_message_finish (msg_io->msg);
 
+        io->in_callback--;
         return 0;
 }
 
