@@ -717,6 +717,14 @@ do_logging_test (Test *test, gconstpointer data)
         g_uri_unref (uri);
 }
 
+static void
+msg_got_body_data_cb (SoupMessage *msg,
+                      guint        chunk_size,
+                      guint64     *response_body_bytes_received)
+{
+        *response_body_bytes_received += chunk_size;
+}
+
 static void
 do_metrics_size_test (Test *test, gconstpointer data)
 {
@@ -724,10 +732,14 @@ do_metrics_size_test (Test *test, gconstpointer data)
         SoupMessage *msg;
         GBytes *response;
         GError *error = NULL;
+        guint64 response_body_bytes_received = 0;
         GBytes *bytes = g_bytes_new_static ("Test", sizeof ("Test"));
 
         uri = g_uri_parse_relative (base_uri, "/echo_post", SOUP_HTTP_URI_FLAGS, NULL);
         msg = soup_message_new_from_uri (SOUP_METHOD_POST, uri);
+        g_signal_connect (msg, "got-body-data",
+                          G_CALLBACK (msg_got_body_data_cb),
+                          &response_body_bytes_received);
         soup_message_set_request_body_from_bytes (msg, "text/plain", bytes);
         soup_message_add_flags (msg, SOUP_MESSAGE_COLLECT_METRICS);
 
@@ -745,6 +757,7 @@ do_metrics_size_test (Test *test, gconstpointer data)
         g_assert_cmpuint (soup_message_metrics_get_response_header_bytes_received (metrics), >, 0);
         g_assert_cmpuint (soup_message_metrics_get_response_body_size (metrics), ==, g_bytes_get_size (response));
         g_assert_cmpuint (soup_message_metrics_get_response_body_bytes_received (metrics), >, soup_message_metrics_get_response_body_size (metrics));
+        g_assert_cmpuint (soup_message_metrics_get_response_body_bytes_received (metrics), ==, response_body_bytes_received);
 
         g_bytes_unref (response);
         g_bytes_unref (bytes);
@@ -1174,6 +1187,7 @@ static gboolean
 unpause_message (SoupServerMessage *msg)
 {
         soup_server_message_unpause (msg);
+        g_object_unref (msg);
         return FALSE;
 }
 
@@ -1200,7 +1214,7 @@ server_handler (SoupServer        *server,
                         soup_server_message_pause (msg);
                         timeout = soup_add_timeout (g_main_context_get_thread_default (),
                                                     is_timeout ? 4000 : 1000,
-                                                    (GSourceFunc)unpause_message, msg);
+                                                    (GSourceFunc)unpause_message, g_object_ref (msg));
                         g_source_unref (timeout);
                 }
         } else if (strcmp (path, "/no-content") == 0) {
